
#include <vector>
#include <match.hh>

class input_module {
public:

  std::string module;

  std::string name, phys;
  uint32_t bustype, vendor, product, version;

  uint32_t matchbits;
  enum { // same order as MATCH_* in input-match.cc !!!
    BITS_EV = 0,
    BITS_KEY,
    BITS_REL,
    BITS_ABS,
    BITS_MSC,
    BITS_LED,
    BITS_SND,
    BITS_FF,

    BITS_LAST
  };

  std::vector<uint32_t> bits[BITS_LAST];

  uint32_t driver_info;
  
  bool operator== (const input_module& other);

  void read_from_sysfs (const std::string& file);
  void read_from_proc (const std::string& file);
  void read_from_env ();
};

std::ostream& operator<< (std::ostream& os, const input_module& module);

class input_subsystem : public basic_subsystem
{
public:
  
  virtual ~input_subsystem  () {}
  
  virtual std::string name () { return "input"; }
  
  virtual void hotplug ();
  virtual void hotplug_all ();
  virtual void list ();
  
private:

  void hotplug (const input_module& m);
  void list (const input_module& m);

  void match_modules (input_module& m,
		      void(input_subsystem::*op)(const input_module&));
  
  void parse_modules (const std::string &fname);
  void parse_modules ();
  void for_each (void(input_subsystem::*op)(const input_module&));
  
  static std::vector<input_module> modules;
};
