
#include <string>
#include <fstream>
#include <iostream>

#include <support.hh>
#include <os.hh>

#include <inttypes.h>

#include "Glob.hh"
#include "isapnp-match.hh"

bool isapnp_module::operator== (const isapnp_module& dev) {
  // FIXME check
  const uint16_t any = uint16_t(-1);
  
  if (cardvendor != any && cardvendor != dev.cardvendor)
    return false;
  if (carddevice != any && carddevice != dev.carddevice)
    return false;

  for(int i=0; i<dev.ndevs; i++) {
    for(int j=0; j<ndevs; j++) {
      if (devs[i].vendor != any && devs[i].vendor != dev.devs[i].vendor)
	continue;
      if (devs[i].function != any && devs[i].function != dev.devs[i].function)
	continue;

      return true;
    }
  }
  
  return false;
};

void isapnp_module::read_from_sysfs (const std::string& file) {
  std::ifstream f (file.c_str());
  char buf[8];
  unsigned int x;

  // read card vendor/device
  f.getline(buf, 8); 
  cardvendor = ISAPNP_VENDOR(buf[0], buf[1], buf[2]);
  sscanf(buf+3, "%4x", &x);
  carddevice = ISAPNP_DEVICE(x);

  // read devices
  ndevs = 0;
  while (f.peek() > 0) {
    f.getline(buf, 8); 

    devs[ndevs].vendor = ISAPNP_VENDOR(buf[0], buf[1], buf[2]);
    sscanf(buf, "%4x", &x);
    devs[ndevs].function = ISAPNP_DEVICE(x);

    ndevs++;
  }
}

void isapnp_module::read_from_proc (const std::string& file) {
  //TODO
  //std::ifstream f (file.c_str());
  //while ( (getline < "/proc/bus/isapnp/devices") > 0 ) {
  //            id = substr ($2,0,7);
  ndevs = 0;
}

void isapnp_module::read_from_env () {
  //TODO
  ndevs = 0;
}

std::ostream& operator<< (std::ostream& os, const isapnp_module& m) {
  os.setf (std::ios::hex, std::ios::basefield);
  os << "cardvendor: " << m.cardvendor
     << ", carddevice: " << m.carddevice;

  for(int i=0; i<m.ndevs; i++) {
    os << ", vendor: " << m.devs[i].vendor
       << ", function: " << m.devs[i].function;
  }
  
  if (!m.module.empty())
    os << ": Module: " << m.module;
  
  return os;
}

// ...

std::vector<isapnp_module> isapnp_subsystem::modules;

void isapnp_subsystem::hotplug ()
{
  parse_modules();
  isapnp_module m;
  m.read_from_env ();
  match_modules (m, &isapnp_subsystem::hotplug);
}

void isapnp_subsystem::hotplug (const isapnp_module& m)
{
  if (m.module.size()) {
    modprobe (m.module);
  }
}

void isapnp_subsystem::hotplug_all ()
{
  parse_modules();
  for_each (&isapnp_subsystem::hotplug);
}

void isapnp_subsystem::list ()
{
  std::cout << "ISA Devices and matched modules:" << std::endl;
  parse_modules();
  
  for_each (&isapnp_subsystem::list);
  std::cout << std::endl;
}

void isapnp_subsystem::list (const isapnp_module& m)
{
  std::cout << m << std::endl;
}

void isapnp_subsystem::match_modules (isapnp_module& m,
				      void(isapnp_subsystem::*op)(const isapnp_module&))
{
  std::vector<isapnp_module>::iterator module = modules.begin();
  while (module != modules.end()) {
    module = std::find (module, modules.end(), m);
    if (module != modules.end()) {
      m.module = module->module;
        (this->*op)(m);
	++module;
    }
  }
}

void isapnp_subsystem::for_each (void(isapnp_subsystem::*op)(const isapnp_module&))
{
  Utility::Glob isapnp_devices ("/sys/bus/pnp/devices/??:??""/id");
  for (Utility::Glob::iterator it = isapnp_devices.begin();
       it != isapnp_devices.end(); ++it)
    {
      isapnp_module m;
      m.read_from_sysfs (*it);
      match_modules (m, op);
    }
  if (isapnp_devices)
    return;

  Utility::Glob isapnp_devices2 ("/proc/bus/isapnp/devices");
  for (Utility::Glob::iterator it = isapnp_devices2.begin();
       it != isapnp_devices2.end(); ++it)
  {
    isapnp_module m;
    m.read_from_proc (*it);
    match_modules (m, op);
  }
}

void isapnp_subsystem::parse_modules (const std::string &fname)
{
  std::ifstream f (fname.c_str());
  if (!f.is_open()) return;

  f.setf (std::ios::hex, std::ios::basefield);

  while (!f.eof() /* FIXME something wrong with eof()? */ && f.peek() > 0) {
    
    // quick hack to skip the leading comment line ...
    if (f.peek() == '#') {
      std::string temp;
      std::getline (f, temp);
      continue;
    }
    
    isapnp_module m;
    
    f >> m.module >> m.cardvendor >> m.carddevice >> m.driver_data;

    // vendor function pairs
    // HACK alert
    char buf[256];
    f.getline(buf, 256);
    m.ndevs = 0;

    char *ptr = buf;
    unsigned int a,b;

    while (*ptr) {
      while (*ptr && ((*ptr == ' ') || (*ptr == '\t'))) ptr++;
    
      if (sscanf(ptr, "%x %x", &a, &b) == 2) {
	m.devs[m.ndevs].vendor = a;
	m.devs[m.ndevs].function = b;
	m.ndevs++;
      }

      while (*ptr && ((*ptr != ' ') && (*ptr != '\t'))) ptr++;
      while (*ptr && ((*ptr == ' ') || (*ptr == '\t'))) ptr++;
      while (*ptr && ((*ptr != ' ') && (*ptr != '\t'))) ptr++;
    }

    //std::cout << m << std::endl;
    
    if (f.good())
      modules.push_back (m);
    
    if (f.fail() && !f.eof())
      f.clear (f.rdstate() & ~std::ios::failbit);
  }
}

void isapnp_subsystem::parse_modules ()
{
  if (modules.size())
    return;
  
  parse_modules("/lib/modules/" + kernel_release() + "/modules.isapnpmap");
  Utility::Glob maps("/etc/hotplug/pnp.*map");
  for (Utility::Glob::iterator it = maps.begin(); it != maps.end(); ++it)
    parse_modules(*it);
 }
