
#include <vector>
#include <match.hh>

#define ISAPNP_CARD_DEVS	8
#define ISAPNP_VENDOR(a,b,c)    (((((a)-'A'+1)&0x3f)<<2)|\
                                ((((b)-'A'+1)&0x18)>>3)|((((b)-'A'+1)&7)<<13)|\
                                ((((c)-'A'+1)&0x1f)<<8))
#define ISAPNP_DEVICE(x)        ((((x)&0xf000)>>8)|\
                                 (((x)&0x0f00)>>8)|\
                                 (((x)&0x00f0)<<8)|\
                                 (((x)&0x000f)<<8))

class isapnp_module {
public:
  std::string module;

  uint16_t cardvendor, carddevice;

  struct {
    unsigned short vendor, function;
  } devs[ISAPNP_CARD_DEVS];       /* logical devices */
  int ndevs;

  uint32_t driver_data;
  
  bool operator== (const isapnp_module& other);

  void read_from_sysfs (const std::string& file);
  void read_from_proc (const std::string& file);
  void read_from_env ();
};

std::ostream& operator<< (std::ostream& os, const isapnp_module& module);

class isapnp_subsystem : public basic_subsystem
{
public:
  
  virtual ~isapnp_subsystem  () {}
  
  virtual std::string name () { return "isapnp"; }
  
  virtual void hotplug ();
  virtual void hotplug_all ();
  virtual void list ();
  
private:

  void hotplug (const isapnp_module& m);
  void list (const isapnp_module& m);

  void match_modules (isapnp_module& m,
		      void(isapnp_subsystem::*op)(const isapnp_module&));
  
  void parse_modules (const std::string &fname);
  void parse_modules ();
  void for_each (void(isapnp_subsystem::*op)(const isapnp_module&));
  
  static std::vector<isapnp_module> modules;
};
