
#include <vector>
#include <match.hh>

class usb_module {
public:
  
  std::string module;
  unsigned int match_flags, idVendor, idProduct, bcdDevice_lo, bcdDevice_hi,
               bDeviceClass, bDeviceSubClass, bDeviceProtocol, bInterfaceClass,
               bInterfaceSubClass, bInterfaceProtocol, driver_info;

  bool operator== (const usb_module& other);
  
  void read_from_sysfs (const std::string& file);
  void read_from_proc (const std::string& file);
  void read_from_env ();
};

std::ostream& operator<< (std::ostream& os, const usb_module& m);

class usb_subsystem : public basic_subsystem
{
public:
  
  virtual ~usb_subsystem  () {}
  
  virtual std::string name () { return "usb"; }
  
  virtual void hotplug ();
  virtual void hotplug_all ();
  virtual void list ();
  
private:

  void hotplug (const usb_module& m);
  void list (const usb_module& m);


  void match_modules (usb_module& m,
		      void(usb_subsystem::*op)(const usb_module&));
  
  void parse_modules (const std::string &fname);
  void parse_modules ();
  void for_each (void(usb_subsystem::*op)(const usb_module&));
  
  static std::vector<usb_module> modules;
};
